/*
Copyright (C) 2009 -  Claudio Salvadori and Christian Nastasi
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/*
 * hv7131gp_hal.c
 *
 * Created on: 9-feb-2009
 * Author: Claudio Salvadori
 */

#include "hv7131gp_hal.h"
#include "hv7131gp_time_out_def.h"

/******************************************************************************/
/*                         Global Variables                                   */
/******************************************************************************/

extern volatile uint8_t *frame_buffer;

extern volatile void (* capture_complete_func) (hv7131gp_status_t);
extern volatile void (* timeout_func) (void);

#ifdef Debug
extern volatile unsigned int d1, d2;
#endif

/******************************************************************************/
/*                          Local Variables                                   */
/******************************************************************************/

volatile uint8_t  row_id = 0;
volatile uint16_t frame_idx = 0;

volatile uint8_t  height;
volatile uint8_t  width;
volatile uint16_t size;

/******************************************************************************/
/*                           Hardware Abstraction Layer                       */
/******************************************************************************/
/* ----------------------------------------------------------------------------
|  Start OC1 Interface:                                                        |
|                                                                              |
|  OC1: Generate MasterClock for Sparkfun Camera (use Timer3)                  |
|       - OC in Toggle Mode -> F_MasterClock = F_TMR/2                         |
 ---------------------------------------------------------------------------- */
void
oc1_init(uint16_t frequency)
{

	/* TIMER 3 for OC */
	T3CON = 0;			/* Stop Timer3 and reset control reg */
	IEC0bits.T3IE = 0;		/* Disable Timer3 interrupts */
	IFS0bits.T3IF = 0;		/* Reset Timer3 interrupt flag */
	TMR3 = 0;			/* Clear timer register */
	PR3 = 0x01;			/* Period -> F_TMR = 20MHz @ 40MHz */

	TRISDbits.TRISD0 = 0;		/* Set OC1 as output */
	OC1CONbits.OCM = 3;		/* Toggle Mode */
	OC1CONbits.OCTSEL = 1;		/* Select Timer3 */
	OC1R  = 0x00;			/* Load OCR */
}



/* ----------------------------------------------------------------------------
|  Init timer5 for timeout:                                                    |
|                                                                              |
|  Initializes the Timer5 						       |
 ---------------------------------------------------------------------------- */
void
timer5_timeout_init(void)
{
	T5CON = 0;		/* Stops the Timer1 and reset control reg */
	TMR5 = 0;		/* Clear contents of the timer register	*/
	IPC7bits.T5IP = 5;	/* Set Timer5 priority to 5 */
}

/* ----------------------------------------------------------------------------
|  Start timer5 for timeout:                                                   |
|                                                                              |
|  Start the timeout with the specific timer setting during init function      |
 ---------------------------------------------------------------------------- */
void
timer5_timeout_set(hv7131gp_time_t timeout)
{
	T5CONbits.TON = 0;		/* Stop Timer5 */
	TMR5 = 0;			/* Clear contents of the timer register	*/
	IFS1bits.T5IF = 0;		/* Clear the Timer5 interrupt status flag */
	IEC1bits.T5IE = 1;		/* Enable Timer5 interrupts */
	switch (timeout) {
	  case HV7131GP_TIME_MORE_THAN_1ms :
		PR5 = 0xEA60;		/* 1.5ms @ FCY=40MHhz*/
		T5CONbits.TCKPS = 0;
		break;
	  case HV7131GP_TIME_MORE_THAN_4_MCLK :
		PR5 = 0x20; 		/* 4MCLK=0x10 @ FCY=40MHhz,MCLK=10MHz */
		T5CONbits.TCKPS = 0;
		break;
	  case HV7131GP_TIME_2086000_MCLK :
		PR5 = 0x7F52; 		/* @ FCY=40MHhz and MCLK = 10MHz */
		T5CONbits.TCKPS = 3;
		break;
	}

	T5CONbits.TON = 1;		/* Start Timer5 without prescaler */
}

/* ----------------------------------------------------------------------------
|  ISR2 Timer5:                                                                |
|                                                                              |
|  Interrupt that manages the init function.                                   |
 ---------------------------------------------------------------------------- */
ISR2(_T5Interrupt)
{
	T5CONbits.TON = 0;	/* Stop Timer5 */
	IEC1bits.T5IE = 0;	/* Disable Timer5 interrupts */

	timeout_func();		/* Call timeout callback  */
	IFS1bits.T5IF = 0;	/* Clear interrupt flag */
}

/* ----------------------------------------------------------------------------
|  ISR2 Timer5:                                                                |
|                                                                              |
|  Interrupt that menages the init function.                                   |
 ---------------------------------------------------------------------------- */
ISR2(_INT0Interrupt)
{
	if (capture_complete_func)
		capture_complete_func(HV7131GP_SUCCESS);
	IFS0bits.INT0IF = 0;	/* Clear interrupt flag */
}

/* ----------------------------------------------------------------------------
|  ISR1 CN:                                                                    |
|                                                                              |
|  New row.                                                                    |
|  Change notify interrupt to understand the start of a new row; that means    |
|  the pixel value after this interrupt are valid value. So what this interrupt|
|  does is the pixel interrupt activation (start of row) and pixel interrupt   |
|  deactivation (end of row).                                                  |
|  In this interrupt is is implemented a procedure to resync the image buffer  |
|  if some pixels are lost in the previous row.                                |
|  This interrupt manages also the end-of-frame notification, switching on     |
|  an interrupt (INT0) that call a callback function                           |
 ---------------------------------------------------------------------------- */
void __attribute__((interrupt)) _CNInterrupt(void)
{
	int rem = 0;

	if (PORTDbits.RD6 == 1) {
		HV7131GP_PIN_VCLK_START();
		rem = frame_idx % width;
		if(rem != 0)
			frame_idx += width - rem; //resync after pixel lost
	}
	else {
		HV7131GP_PIN_VCLK_STOP();
		/* Check if last row is ready -> frame ready? */
		if ((frame_idx >= size) || (++row_id >= height)){
			/* Stop row and frame syncs, notify capture complete */
			HV7131GP_PIN_HSYNC_STOP();
			HV7131GP_PIN_VSYNC_STOP();
			IFS0bits.INT0IF = 1; //Activate interrupt
		}
	}

	IFS1bits.CNIF = 0;			/* Reset CN interrupt flag */
}

/* ----------------------------------------------------------------------------
|  ISR1 INT3:                                                                  |
|                                                                              |
|  New Frame.                                                                  |
|                                                                              |
|  Zeroing of row and pixel counters, set image dimension variables, start     |
|  reception of row interrupts.                                                |
|  If "Debug" configuration is possible to manage the function that measure    |
|  the frame duration, and toggle a value of output pin to show that duration  |
 ---------------------------------------------------------------------------- */
void __attribute__((interrupt)) _INT3Interrupt(void)
{

#ifdef Debug
	timer32_end();
	LATGbits.LATG1 ^= 1; 		//This pin is toggled every frame
	timer32_start();
#endif //Debug

	row_id 		= 0;
	frame_idx 	= 0;

	//Read the size of the picture
	width 	= hv7131gp_get_width();
	height 	= hv7131gp_get_height();
	size 	= hv7131gp_get_size();

	HV7131GP_PIN_HSYNC_START();

	HV7131GP_PIN_VSYNC_IF_RESET();    /* Reset INT3 interrupt flag */
}

/* ----------------------------------------------------------------------------
|  ISR1 INT4:                                                                  |
|                                                                              |
|  Video (pixels) clock interrupt.                                             |
|  Fast interrupt for acquisition of the 8 bit of the single pixel.            |
|  Here is also updated the pixel counter.                                     |
|  With MCLK = 10 MHz and time divisor = 1 and 2 is possible to lost some      |
|  pixel (less than 1 for line); in the row interrupt is implemented a         |
|  a procedure of recovery sync caused by pixel lost                           |
 ---------------------------------------------------------------------------- */
void __attribute__((interrupt, shadow, no_auto_psv)) _INT4Interrupt(void)
{
	IFS3bits.INT4IF = 0;    		/* Reset INT4 interrupt flag */

	frame_buffer[frame_idx]=((uint8_t) PORTE)
	                        & 0x00FF; 	//Pixel acquisition
	frame_idx++; 				//Pixel counter update
}
/*******************************************************************************
* 				Debug functions                                *
*******************************************************************************/
#ifdef Debug
/* ----------------------------------------------------------------------------
|  Init timer 8/9 for frame duration:                                          |
|                                                                              |
|  Initializes the Timer8/9 and set the timeout parameters                     |
 ---------------------------------------------------------------------------- */
void
timer32_init(void)
{
	T8CON = 0;		/* Stops the Timer2 and reset control reg */
	T8CONbits.T32 = 1; 	//32Bits timer
	PR9 = 0xFFFF;   	/* Set timer to the maximum, avoiding interrupt*/
	PR8 = 0xFFFF;
	IEC3bits.T9IE = 0; 	//Interrupt disabled
	IFS3bits.T9IF = 0; 	//Clean interrupt flag
}

/* ----------------------------------------------------------------------------
|  Start timer 8/9 (32bits) for frame duration                                 |
|                                                                              |
|  Start Timer 8/9 and variables zeroing                                       |
 ---------------------------------------------------------------------------- */
void
timer32_start(void)
{
	TMR8 = 0;		/* Clear contents of the timer register	*/
	TMR9 = 0;		/* Clear contents of the timer register	*/
	T8CONbits.TON = 1; 	//Start timer
}

/* ----------------------------------------------------------------------------
|  Stop timer8/9 (32bits) for frame duration                                   |
|                                                                              |
|  Stop timer 8/9 and update variables                                         |
 ---------------------------------------------------------------------------- */
void
timer32_end(void)
{
	d2 = TMR8;		/* Save contents of the timer register	*/
	d1 = TMR9;		/* Save contents of the timer register	*/
	T8CONbits.TON = 0;	 //Start timer
}
#endif
